#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>		// For write()
#include <unistd.h>
#include <sys/socket.h>		// For connect()
#include <string.h>		// For memset()
#include <fcntl.h>		// For fcntl()
#include <sys/time.h>		// For select()
#include <netinet/in.h>		// For
#include <arpa/inet.h>		// inet_addr()

#include <client.h>
#include <main.h>
#include <conf.h>
#include <logger.h>
#include <queue.h>
#include <server.h>
#include <globals.h>

// Check if there are clients trying to connect
// Add new clients if gossip has room
void
handle_incoming(void)
{
    int             tempsocket;
    struct sockaddr_in their_addr;
    int             i,
                    sin_size,
                    no_connects;

    sin_size = sizeof(struct sockaddr_in);

    if ((tempsocket = accept(sockets[1], (struct sockaddr *)&their_addr, &sin_size)) >= 0) {
	// We got a new connection
	// Check if limit of connections from this IP-address has been reached
	no_connects = 0;
	for (i = 0; i < MAXADMINS; i++) {
	    if (clients[i].socket >= 0 && !strcmp(clients[i].ip, inet_ntoa(their_addr.sin_addr))) {
		no_connects++;
	    }
	}
	if (no_connects >= conf_getint("connects_per_ip")) {
	    // Connection limit for this IP-address reached, disconnect
	    do_log(LOGGER_WARNING,
		   "Connection from %s aborted: connection-limit reached", inet_ntoa(their_addr.sin_addr));
	    close(tempsocket);
	    return;
	}

	for (i = 0; i < MAXADMINS; i++) {
	    // Find a place to stash this socket
	    if (clients[i].socket == -1) {
		clients[i].socket = tempsocket;
		clients[i].loggedin = 0;
		clients[i].relaylogs = 0;
		clients[i].logintime = time(NULL);
		clients[i].lasttime = clients[i].logintime;
		strcpy(clients[i].ip, inet_ntoa(their_addr.sin_addr));
		tempsocket = -1;	// Ok, we stashed it
		break;
	    }
	}
	// No room for this socket?
	if (tempsocket != -1) {
	    // Sorry, we need to disconnect. A warning could be a good idea for the future though..
	    do_log(LOGGER_WARNING, "Connection from %s aborted: GOSSIP full", inet_ntoa(their_addr.sin_addr));
	    close(tempsocket);
	    return;
	}
	do_log(LOGGER_INFO, "New connection from %s, assigned client-number %d", inet_ntoa(their_addr.sin_addr), i);
    }
}

// Check if there's data from the clients. If so: send it to the server
// and add client-handle to the queue
void
handle_clientdata(void)
{
    fd_set          readfds;
    int             counter;
    int             maxsocket = 0;
    struct timeval  tv;
    char            readbuf[BUFSIZ];
    int             bytesread,
                    retval,
                    clientfd;

    // Add living clients to fdset
    FD_ZERO(&readfds);
    for (counter = 0; counter < MAXADMINS; counter++) {
	clientfd = clients[counter].socket;
	if (clientfd != -1)
	    FD_SET(clientfd, &readfds);
	if (clientfd > maxsocket)
	    maxsocket = clientfd;
    }

    // Check if there's incoming data
    tv.tv_sec = 0;
    tv.tv_usec = 20000;

    // When select returns 0 on non-blocking sockets, the fdset isn't
    // altered, so FD_ISSET will return true although there's no data
    // to be read. Back out here.
    if ((retval = select(maxsocket + 1, &readfds, NULL, NULL, &tv)) <= 0)
	return;

    for (counter = 0; counter < MAXADMINS; counter++) {
	clientfd = clients[counter].socket;

	if (clientfd != -1 && FD_ISSET(clientfd, &readfds)) {
	    if ((bytesread = read(clientfd, readbuf, sizeof(readbuf))) > 0) {
		clients[counter].lasttime = time(NULL);
		if (clients[counter].loggedin == 0)
		    handle_login(counter, readbuf, bytesread);
		else {
		    if (!handle_gossipcommand(counter, readbuf, bytesread)) {
			send_clientdata(counter, readbuf, bytesread);
		    }
		}
	    } else {
		// Error reading, close this client connect and clean up
		do_log(LOGGER_DEBUG, "Client connect %d died in handle_clientdata(): %s", counter, strerror(errno));
		drop_client(counter);
	    }
	}
    }
}

// Check if any client timed out, if so: disconnect it
void
handle_timeouts(void)
{
    int             i;

    for (i = 0; i < MAXADMINS; i++) {
	if (clients[i].socket != -1) {
	    if ((clients[i].loggedin == 0 && time(NULL) >= clients[i].logintime + conf_getint("login_timeout"))
		|| (clients[i].loggedin == 1 && time(NULL) >= clients[i].lasttime + conf_getint("inactivity_timeout"))) {
		drop_client(i);
		do_log(LOGGER_INFO, "Client %d (%s) logged out due to timeout", i, clients[i].ip);
	    }
	}
    }
}

// Send data from the server to the right client
void
send_serverdata(char *buf, int bufsize)
{
    fd_set          writefds;
    struct timeval  tv;
    int             client,
                    clientfd,
                    retval = 0;
    int             selecttries = 0;

    if (queue_empty() || queue[queue_start] < 0)
	return;		// Discard data if there's no client waiting for it


    tv.tv_sec = 0;
    tv.tv_usec = 20000;
    FD_ZERO(&writefds);

    client = queue[queue_start];
    clientfd = (clients[client]).socket;
    FD_SET(clientfd, &writefds);

    // Try to do the select a few times
    while ((retval = select(clientfd + 1, NULL, &writefds, NULL, &tv)) == 0) {
	selecttries++;
	if (selecttries == 3)
	    break;
    }

    if (FD_ISSET(clientfd, &writefds)) {
	if (write(clientfd, buf, bufsize) < 0) {
	    // Connection dead, close socket and clean up
	    do_log(LOGGER_DEBUG, "Client %d (%s) died in send_serverdata()", client, clients[client].ip);
	    drop_client(client);
	}
    } else {
	// Could not write, too bad
	do_log(LOGGER_DEBUG,
	       "Client %d (%s) not ready to receive data in send_serverdata()", client, clients[client].ip);
    }
}


// Relay data to all clients
void
do_broadcast(char *buf, int bufsize)
{
    fd_set          writefds;
    struct timeval  tv;
    int             i,
                    retval;
    int             maxsocket = -1;
    int             selecttries = 0;

    // Initialize for select()
    tv.tv_sec = 0;
    tv.tv_usec = 20000;
    FD_ZERO(&writefds);

    // Add active clients to fd_set
    for (i = 0; i < MAXADMINS; i++) {
	if (clients[i].socket != -1 && clients[i].relaylogs == 1) {
	    FD_SET(clients[i].socket, &writefds);
	    if (clients[i].socket > maxsocket)
		maxsocket = clients[i].socket;
	}
    }

    if (maxsocket == -1)
	return;			// No clients

    // Find clients which are ready to receive and send them the data
    // Will try to do a select a few times. If select keeps failing, just
    // write to the socket and hope things work out..
    while ((retval = select(maxsocket + 1, NULL, &writefds, NULL, &tv)) == 0) {
	selecttries++;
	if (selecttries == 3)
	    break;
    }

    for (i = 0; i < MAXADMINS; i++) {
	if (clients[i].socket != -1 && FD_ISSET(clients[i].socket, &writefds)) {
	    if (write(clients[i].socket, buf, bufsize) < 0) {
		// Connection dead, close socket and clean up
		do_log(LOGGER_DEBUG, "Client %d (%s) died in do_broadcast()", i, clients[i].ip);
		drop_client(i);
	    }
	}
    }
}

// Let client login to GOSSIP
void
handle_login(int client, char *buf, int bufsize)
{
    int             pwdlen = 0;
    int             datalen = 0;
    char            pwd[256];
    char            data[32];
    char            errstr[1024];

    if (strlen(buf) < (strlen("hello,connect") + 2)) {
	// Can't be a valid login, clean up, drop connection
	do_log(LOGGER_WARNING, "Invalid login attempt from client %d (%s)", client, clients[client].ip);
	drop_client(client);
	return;
    }

    pwdlen = (int)buf[0];
    if (pwdlen > (strlen(buf) - strlen("hello,connect") - 2)) {
	// Can't be a valid login
	do_log(LOGGER_WARNING, "Invalid login attempt from client %d (%s)", client, clients[client].ip);
	drop_client(client);
	return;
    }

    strncpy(pwd, buf + 1, pwdlen);
    pwd[pwdlen] = '\0';
    datalen = (int)buf[1 + pwdlen];

    if (datalen != strlen("hello,connect")
	|| strlen(buf) != (pwdlen + 1 + strlen("hello,connect") + 1)) {
	// Can't be a valid login
	do_log(LOGGER_WARNING, "Invalid login attempt from client %d (%s)", client, clients[client].ip);
	drop_client(client);
	return;
    }

    strncpy(data, buf + 1 + pwdlen + 1, strlen("hello,connect"));
    data[datalen] = '\0';

    // Ok, we should have the password in pwd and "hello,connect" in data
    // let's check.
    if (strcmp(data, "hello,connect")) {
	do_log(LOGGER_WARNING, "Invalid login attempt from client %d (%s)", client, clients[client].ip);
	drop_client(client);
	return;
    }
    if (strcmp(pwd, (char *)conf_get("adminpass"))) {
	sprintf(errstr, "PASSWORD:Invalid password %s!", pwd);
	write(clients[client].socket, errstr, strlen(errstr));
	do_log(LOGGER_WARNING, "Password \"%s\" from client %d (%s) incorrect", pwd, client, clients[client].ip);
	drop_client(client);
    } else {
	snprintf(errstr, sizeof(errstr), "COMMAND hello,connect\nexecuted on the server!\n");
	write(clients[client].socket, errstr, strlen(errstr));
	clients[client].loggedin = 1;
	do_log(LOGGER_DEBUG, "Client %d (%s) logged in", client, clients[client].ip);
    }

    return;
}

// Disconnect client, clean stuff up
void
drop_client(int client)
{
    int             i;

    // Scrap client from the queue
    for (i = queue_start; i <= queue_end; i++) {
	if (queue[i] == client)
	    queue[i] = -1;
    }

    close(clients[client].socket);	// Close socket
    clients[client].socket = -1;
    do_log(LOGGER_INFO, "Client %d (%s) disconnected", client, clients[client].ip);
}
