#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <conf.h>
#include <misc.h>

#include <globals.h>


// Read and parse configuration file. Fill in the configuration-template
// and return success or failure
int
config(void)
{
    char           *cfile;
    char            tempstr[256];
    int             counter = 0;
    int             retval = 0;
    FILE           *fp;

    // Override default filename with command-line option
    if (configfile)
	cfile = configfile;
    else
	cfile = CONFIGFILE;

    // Open configuration file
    if ((fp = fopen(cfile, "r")) == NULL) {
	printf("Could not open configuration file %s: %s\n", cfile, strerror(errno));
	return -1;
    }
    // Read and parse configuration file
    while (fgets(tempstr, sizeof(tempstr), fp) != NULL) {
	counter++;
	strcrop(tempstr);
	if (strlen(tempstr) == 0)
	    continue;		// Empty line, ignore 
	if (tempstr[0] == '#')
	    continue;		// Comment, ignore

	if (conf_parseline(tempstr) < 0) {
	    printf("Error parsing configfile on line %d: %s\n", counter, tempstr);
	    fclose(fp);
	    return -1;
	}
    }

    if (ferror(fp)) {
	printf("Error reading configuration file: %s\n", strerror(errno));
	return -1;
    }

    fclose(fp);

    counter = 0;
    while (strcmp(configuration[counter].name, "")) {
	if (configuration[counter].needed == CONF_NEEDED && configuration[counter].value == NULL) {
	    printf("Configurable %s needed but not configured\n", configuration[counter].name);
	    retval = -1;
	}
	counter++;
    }
    return retval;
}




// Parse configuration-string, fill in configuration template
// Returns 0 if successful, -1 if failed 
int
conf_parseline(char *str)
{
    char            tmpname[256];
    char            tmpvalue[256];
    int             stored_value,
                    i;

    if (sscanf(str, "%255[^ =]%*[ =]%255s", tmpname, tmpvalue) < 2) {
	return -1;
    }
    strcrop(tmpname);
    strcrop(tmpvalue);
    locase(tmpname);

    stored_value = 0;
    i = 0;
    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, tmpname)) {
	    if (configuration[i].value != NULL) {
		stored_value = 1;
		break;
	    }
	    switch (configuration[i].eltype) {
	    case CONF_ELTYPE_INT:{
		    conf_setint(configuration[i].name, atoi(tmpvalue));
		    break;
		}
	    case CONF_ELTYPE_STRING:{
		    conf_setstring(configuration[i].name, tmpvalue);
		    break;
		}

	    case CONF_ELTYPE_YESNO:{
		    if ((!strcmp(locase(tmpvalue), "yes")) || (!strcmp(locase(tmpvalue), "no"))) {
			if (!strcmp(locase(tmpvalue), "yes")) {
			    conf_setint(configuration[i].name, 1);
			}
			if (!strcmp(locase(tmpvalue), "no")) {
			    conf_setint(configuration[i].name, 0);
			}
		    } else {
			return -1;
		    }
		    break;
		}
	    default:
		break;
	    }
	    stored_value = 1;
	}
	i++;
    }

    if (!stored_value) {
	printf("Unknown configurable: %s\n", tmpname);
	return -1;
    }
    return 0;
}


// Return pointer to value of configurable 'name' or NULL if not found
void           *
conf_get(char *name)
{
    int             i;

    i = 0;
    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, name))
	    return configuration[i].value;
	i++;
    }
    return NULL;
}

// Set a configurable 'name' to 'value'. 
int
conf_setstring(char *name, void *value)
{
    int             i = 0;

    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, name)) {
	    if (configuration[i].value != NULL) {
		free(configuration[i].value);
	    }
	    switch (configuration[i].eltype) {
	    case CONF_ELTYPE_INT:{
		    return -1;
		    break;
		}
	    case CONF_ELTYPE_STRING:{
		    configuration[i].value = malloc(strlen(value) + 1);
		    strcpy(configuration[i].value, value);
		    break;
		}
	    case CONF_ELTYPE_YESNO:{
		    return -1;
		    break;
		}
	    default:
		break;
	    }
	    // Value set.
	    return 0;
	}
	i++;
    }

    // Configureable not found
    return -1;
}


// Set a configurable 'name' to 'value'. 
int
conf_setint(char *name, int value)
{
    int             i = 0;

    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, name)) {
	    if (configuration[i].value != NULL) {
		free(configuration[i].value);
	    }
	    switch (configuration[i].eltype) {
	    case CONF_ELTYPE_INT:{
		    configuration[i].value = malloc(sizeof(int));
		    *((int *)configuration[i].value) = value;
		    break;
		}
	    case CONF_ELTYPE_STRING:{
		    return -1;
		    break;
		}
	    case CONF_ELTYPE_YESNO:{
		    if (value == 0 || value == 1) {
			configuration[i].value = malloc(sizeof(int));
			*((int *)configuration[i].value) = value;
		    } else {
			return -1;
		    }
		    break;
		}
	    default:
		break;
	    }
	    // Value set.
	    return 0;
	}
	i++;
    }

    // Configureable not found
    return -1;
}

// Return 1 if setting was 'yes', 0 if 'no' or value non-existant
int
conf_isyes(char *name)
{
    int             i;

    i = 0;
    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, name))
	    return (*(int *)configuration[i].value);
	i++;
    }
    return 0;
}

// Return the value of 'name' as int
int
conf_getint(char *name)
{
    int             i,
                    retval;

    i = 0;
    retval = 0;
    while (strcmp(configuration[i].name, "")) {
	if (!strcmp(configuration[i].name, name)) {
	    retval = *((int *)(configuration[i].value));
	    break;
	}
	i++;
    }
    return retval;
}
