// GOSSIP - The Global Ops admin tool thingy

#define GOSSIP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>

#include <main.h>
#include <conf.h>
#include <logger.h>
#include <version.h>
#include <queue.h>
#include <misc.h>
#include <server.h>
#include <client.h>

#define LOCAL
#include <globals.h>

int
main(int argc, char **argv)
{
    int             c,
                    tempint;
    unsigned int    sleeptime;

    while ((c = getopt(argc, argv, "c:dDvhl:nL:")) != -1) {
	switch (c) {
	case 'c':
	    configfile = optarg;
	    break;
	case 'd':
	    conf_setint("daemonize", 1);
	    break;
	case 'D':
	    conf_setint("debug", 1);
	    break;
	case 'v':
	    version();
	    break;
	case 'l':
	    conf_setstring("logfilename", optarg);
	    break;
	case 'n':
	    conf_setint("logrelay", 0);
	    break;
	case 'L':
	    tempint = atoi(optarg);
	    if (tempint < 0 || tempint > 3)
		usage(argv[0]);
	    conf_setint("logmethod", tempint);
	    break;
	case '?':
	case 'h':
	default:
	    usage(argv[0]);
	    break;
	}
    }


    // Init stuff
    init();

    // Main loop
    while (1) {
	while (!connected()) {
	    // Reconnect, else wait. Clientdata will not be read!

	    if (connecttogo() < 0) {
		sleeptime = CONNECT_RETRY_WAITTIME;

		// Sleep gets interrupted by SIGALRM
		while ((sleeptime = sleep(sleeptime)) > 0);
	    }
	}

	// Are there any clients trying to connect? Accept them or
	// disconnect them if GOSSIP is full.
	handle_incoming();

	// Handle server-data
	handle_serverdata();

	// Handle client-data
	handle_clientdata();

	// Handle log-relaying
	if (conf_isyes("logrelay"))
	    handle_logrelay();

	// Handle timeouts
	handle_timeouts();

    }

    exit(1);
}

void
init(void)
{
    int             i;


    // Parse config file
    if (config() < 0) {
	printf("Error while parsing configurationfile, program exiting\n");
	exit(1);
    }
    // Initialize logger
    if (logger_init() < 0) {
	printf("Logger initialization failed, program exiting\n");
	exit(1);
    }
    // Initialize queue
    init_queue();

    // Initialize banlist
    // init_banlist();

    // Set sockets to inactive
    sockets[0] = -1;
    sockets[1] = -1;

    // Initialize client-info
    for (i = 0; i < MAXADMINS; i++) {
	clients[i].socket = -1;
    }

    if (do_listen()) {
	do_log(LOGGER_ERROR, "Starting listener failed, gossip exiting");
	exit(1);
    }

    if (conf_isyes("daemonize")) {
	if (daemonize() < 0) {
	    do_log(LOGGER_ERROR, "Could not daemonize, program exiting\n");
	    exit(1);
	}
    }
    // Turn logfile checking off
    do_checklogfiles = 0;

    // If we're going to relay logs, set a signal handler
    if (conf_isyes("logrelay")) {
	setupalarm();
    }
}


// Every now and then a alarm-signal is sent and this function is called
// By setting do_checklogfiles to 1, which is checked in handle_logrelay()
// If set, the newest logfile will be used.
void
handle_alarm(int sig)
{
    do_checklogfiles = 1;
}


// Start a listener for the clients
int
do_listen(void)
{
    struct sockaddr_in my_addr;
    int             yes = 1;

    if ((sockets[1] = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	do_log(LOGGER_ERROR, "Could not create listen socket: %s", strerror(errno));
	return -1;
    }
    // Make address reusable
    setsockopt(sockets[1], SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));

    // Set socket to non-blocking
    fcntl(sockets[1], F_SETFL, O_NONBLOCK);

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(conf_getint("listenport"));
    my_addr.sin_addr.s_addr = INADDR_ANY;
    memset(&(my_addr.sin_zero), '\0', 8);	// zero rest of struct

    if (bind(sockets[1], (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
	do_log(LOGGER_ERROR, "Could not bind to port %d: %s", conf_getint("listenport"), strerror(errno));
	return -1;
    }

    if (listen(sockets[1], MAXADMINS) == -1) {
	do_log(LOGGER_ERROR, "Could not listen: %s", strerror(errno));
	return -1;
    }

    do_log(LOGGER_DEBUG, "Listening on port %d", conf_getint("listenport"));
    return 0;
}


// Checks if the buffer contains a gossip-command and handles it
// Returns true if it was a GOSSIP-command, false if not.
int
handle_gossipcommand(int client, char *buf, int bufsize)
{
    char            tempstr[64];
    int             retval = 0;
    char           *tempptr;

    if (bufsize > sizeof(tempstr) - 1)
	return 0;		// It's too big, can't be a GOSSIP-command

    strncpy(tempstr, buf, bufsize);

    if (!strcmp(tempstr, "GOSSIP|start")) {
	if (!conf_isyes("logrelay")) {
	    do_log(LOGGER_WARNING, "Client %d (%s) tried to start log-relaying, but it's disabled in the config file",
		   client, clients[client].ip);
	    retval = 1;
	} else {
	    do_log(LOGGER_DEBUG, "Client %d (%s) enabled log-relaying", client, clients[client].ip);
	    clients[client].relaylogs = 1;
	    retval = 1;
	}
    }

    if (!strcmp(tempstr, "GOSSIP|stop")) {
	if (!conf_isyes("logrelay")) {
	    do_log(LOGGER_WARNING, "Client %d (%s) tried to stop log-relaying, but it's disabled in the config file",
		   client, clients[client].ip);
	    retval = 1;
	} else {
	    do_log(LOGGER_DEBUG, "Client %d (%s) disabled log-relaying", client, clients[client].ip);
	    clients[client].relaylogs = 0;
	    retval = 1;
	}
    }

    /*
     * if (sscanf(tempstr,"GOSSIP|ban|%s",tempptr)==1) {
     * do_log(LOGGER_DEBUG,"Client %d (%s) name-banned
     * %s",client,clients[client].ip,tempptr); add_to_banlist(tempptr);
     * write_banlist(); retval=1; } 
     */

    return retval;
}


// Relay GO-logs to the clients
void
handle_logrelay(void)
{
    static char     currentlogfile[2048] = "\0";	// Filename of
    							// current logfile
    static FILE    *fp;		// Filehandle for current logfile
    static long     position;	// Current position in the logfile

    char            templogfile[2048];
    struct stat     dirstat;
    DIR            *dir;
    struct dirent   direntry;
    struct dirent  *tempdirentry;
    char            readbuf[BUFSIZ];
    char            writebuf[BUFSIZ + 16];	// To hold the readbuffer
    						// prepended with
    						// "GOSSIP|"
    char            date[10];
    char            count[6];

    // Check logfiles every now and then, if do_checklogfiles is set
    if (do_checklogfiles) {
	do_checklogfiles = 0;

	// Open directory
	if ((dir = opendir(conf_get("logdirectory"))) == NULL) {
	    do_log(LOGGER_WARNING, "Could not open logdirectory %s, logrelaying disabled", conf_get("logdirectory"));
	    conf_setint("logrelay", 0);
	    destroyalarm();
	    return;
	}

	memset(&direntry, 0, sizeof(direntry));	// d_name should be empty

	// Find newest entry in direntry
	while ((tempdirentry = readdir(dir)) != NULL) {
	    // Find newest one..
	    if (sscanf(tempdirentry->d_name, "%8s-%4s.log", date, count) == 2) {
		if (strlen(direntry.d_name) == 0 || isnewer(tempdirentry->d_name, direntry.d_name)) {
		    strcpy(direntry.d_name, tempdirentry->d_name);
		}
	    }
	}

	if (strlen(direntry.d_name) == 0) {
	    // No logfiles in directory
	    closedir(dir);
	    return;
	}

	// Build full filename
	snprintf(templogfile, sizeof(templogfile), "%s/%s", (char *)conf_get("logdirectory"), direntry.d_name);

	// Close directory
	closedir(dir);

	// Stat entry
	if (lstat(templogfile, &dirstat)) {
	    do_log(LOGGER_WARNING, "Could not stat logfile %s: %s", templogfile, strerror(errno));
	    return;
	}

	if (!(dirstat.st_mode && S_IFREG)) {
	    // Logfile is not a file, might be dangerous to read it, so don't.
	    return;
	}

	// Check if it's a new logfile, if so, close old and open new
	if (strcmp(currentlogfile, templogfile)) {
	    if (strlen(currentlogfile) > 0)
		fclose(fp);

	    strncpy(currentlogfile, templogfile, sizeof(currentlogfile));
	    do_log(LOGGER_DEBUG, "Now using logfile %s", currentlogfile);

	    if ((fp = fopen(currentlogfile, "r")) == NULL) {
		do_log(LOGGER_WARNING, "Could not open logfile %s: %s", currentlogfile, strerror(errno));
		return;
	    }
	    fseek(fp, 0, SEEK_END);	// Start at the end of the file,
					// IS THIS OK?
	    position = (long)ftell(fp);
	}
    }

    if (strlen(currentlogfile) == 0)
	return;			// If logfile isn't set, back out

    // Read to the end of the file and relay the read lines
    if (position < (long)dirstat.st_size) {
	while (!feof(fp)) {
	    if (fgets(readbuf, sizeof(readbuf), fp) == 0)
		break;

	    if (ferror(fp)) {
		do_log(LOGGER_WARNING, "Error reading logfile %s: %s", currentlogfile, strerror(errno));
		fclose(fp);
		currentlogfile[0] = '\0';
		return;
	    }
	    // do_namebans(readbuf); 

	    // Relay data
	    // Put "GOSSIP|" in front, then broadcast it to the clients
	    snprintf(writebuf, sizeof(writebuf), "GOSSIP|%s", readbuf);
	    do_broadcast(writebuf, strlen(writebuf));
	}
    }
    position = ftell(fp);	// Save position for next time
    clearerr(fp);		// Clear EOF status
}




// This namebanning stuff isn't completely implemented yet, because GOSA 
// doesn't support it. Too bad BUG-Falcon isn't around much these days..

// Initialize banlist
void
init_banlist(void)
{
    FILE           *fp;
    char            buf[BUFSIZ];
    struct nameban_el *cur_el;

    namebans = NULL;

    if ((fp = fopen((char *)conf_get("banlist"), "r")) == NULL) {
	do_log(LOGGER_ERROR, "Could not open banlist file %s, exiting\n", (char *)conf_get("banlist"));
	exit(1);
    }

    while (!feof(fp)) {
	if (fgets(buf, sizeof(buf), fp)) {
	    if (namebans) {
		cur_el->next = malloc(sizeof(struct nameban_el));
		cur_el = cur_el->next;
	    } else {
		namebans = malloc(sizeof(struct nameban_el));
		cur_el = namebans;
	    }

	    cur_el->name = malloc(strlen(buf) + 1);
	    strcpy(cur_el->name, buf);
	    cur_el->next = NULL;

	}
    }

    fclose(fp);
}

// Add a name to banlist
void
add_to_banlist(char *str)
{
    struct nameban_el *cur_el;

    cur_el = namebans;

    if (!cur_el) {
	cur_el = malloc(sizeof(struct nameban_el));
    } else {
	while (cur_el->next)
	    cur_el = cur_el->next;
	cur_el->next = malloc(sizeof(struct nameban_el));
	cur_el = cur_el->next;
    }
    cur_el->next = NULL;
    cur_el->name = str;
}


// Write banlist to file
void
write_banlist(void)
{
    FILE           *fp;
    struct nameban_el *cur_el;
    char           *enter = "\n";

    if ((fp = fopen((char *)conf_get("banlist"), "w+")) == NULL) {
	do_log(LOGGER_ERROR, "Could not write banlist file %s", (char *)conf_get("banlist"));
	return;
    } else {
	cur_el = namebans;
	while (cur_el) {
	    fputs(cur_el->name, fp);
	    fputs(enter, fp);
	    cur_el = cur_el->next;
	}
    }
    fclose(fp);
}

// Enforce namebans by parsing loglines to see if there's a banned name in it
void
do_namebans(char *str)
{
    // L 01/27/2003 - 20:22:35: "Player<>" entered the game

    char            timedate[64];
    char            name[1024];
    struct nameban_el *cur_el;

    if (strstr(str, "\"Player<>\" entered the game"))
	return;

    if (sscanf(str, "%[^\"]\"%[^\"]%*s", timedate, name) == 2) {
	cur_el = namebans;
	while (cur_el) {
	    if (strstr(name, cur_el->name)) {
		// We got a banned player
		do_log(LOGGER_DEBUG, "Auto-banned player %s", name);
	    }
	    cur_el = cur_el->next;
	}
    }
}

// Guess what this function does
void
usage(char *pgm)
{
    printf("Usage: %s [-dvh] [-l logfile] [-L logmethod]\n", pgm);
    printf("Options:\n");
    printf("\t-c [Conffile]\tUse the given filename as configuration file\n");
    printf("\t-d\t\tEnable daemonizing\n");
    printf("\t-D\t\tEnable debugging\n");
    printf("\t-v\t\tShow version\n");
    printf("\t-h\t\tShow usage (this text)\n");
    printf("\t-l [Logfile]\tUse the given filename as logfile\n");
    printf("\t-n\t\tNo logrelaying\n");
    printf("\t-L [Logmethod]\tLog using the given method:\n");
    printf("\t\t\t0 - Don't log\n");
    printf("\t\t\t1 - Log to syslog only\n");
    printf("\t\t\t2 - Log to file only\n");
    printf("\t\t\t3 - Log to both syslog and file\n");
    printf("\n");
    exit(1);
}

// Somebody set us up the alarm
void
setupalarm(void)
{
    struct itimerval itv;

    // Set alarm
    itv.it_interval.tv_usec = 0;
    itv.it_interval.tv_sec = 2;
    itv.it_value.tv_usec = 0;
    itv.it_value.tv_sec = 2;
    setitimer(ITIMER_REAL, &itv, NULL);
    signal(SIGALRM, handle_alarm);
}

// Somebody didn't set us up the alarm
void
destroyalarm(void)
{
    struct itimerval itv;

    // Destroy alarm
    itv.it_interval.tv_usec = 0;
    itv.it_interval.tv_sec = 0;
    itv.it_value.tv_usec = 0;
    itv.it_value.tv_sec = 0;
    setitimer(ITIMER_REAL, &itv, NULL);
    signal(SIGALRM, SIG_IGN);
}
