#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>		// For write()
#include <unistd.h>
#include <sys/socket.h>		// For connect()
#include <string.h>		// For memset()
#include <fcntl.h>		// For fcntl()
#include <sys/time.h>		// For select()
#include <netinet/in.h>		// For
#include <arpa/inet.h>		// inet_addr()
#include <errno.h>


#include <server.h>
#include <conf.h>
#include <logger.h>
#include <queue.h>
#include <client.h>
#include <globals.h>

// Returns 1 if connected to GO-admin port, 0 if not
int
connected(void)
{
    if (sockets[0] >= 0)
	return 1;
    else
	return 0;
}

// Make a connection to the GO-admin port and log in
// Return -1 to calling function when connecting fails
int
connecttogo(void)
{
    static int      failed = 0;	// To know if we're _re_connecting
    struct sockaddr_in dest_addr;
    char            loginstr[128];
    char            readbuf[BUFSIZ];
    int             retval;

    if ((sockets[0] = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	if (!failed) {
	    do_log(LOGGER_ERROR, "Could not create socket in connecttogo(): %s",strerror(errno));
	}
	close(sockets[0]);
	sockets[0] = -1;
	failed = 1;
	return -1;
    }

    // Set up the sockaddr struct
    memset(readbuf, '\0', sizeof(readbuf));
    dest_addr.sin_family = AF_INET;
    dest_addr.sin_port = htons(conf_getint("adminport"));
    dest_addr.sin_addr.s_addr = inet_addr((char *)conf_get("adminaddress"));
    memset(&(dest_addr.sin_zero), '\0', 8);

    if (connect(sockets[0], (struct sockaddr *)&dest_addr, sizeof(struct sockaddr_in))) {
	if (!failed) {
	    do_log(LOGGER_ERROR, "Could not connect to GO: %s", strerror(errno));
	}
	close(sockets[0]);
	sockets[0] = -1;
	failed = 1;
	return -1;
    }


    // Send login information
    snprintf(loginstr, sizeof(loginstr), "%c%s%chello,connect",
	     strlen((char *)conf_get("adminpass")), (char *)conf_get("adminpass"), strlen("hello,connect"));

    if (write(sockets[0], loginstr, strlen(loginstr)) < 0) {
	if (!failed) {
	    do_log(LOGGER_ERROR, "Could not send login information to GO: %s",strerror(errno));
	} 
	close(sockets[0]);
	sockets[0] = -1;
	failed = 1;
	return -1;
    }

    // Now read reply
    if (read(sockets[0], readbuf, sizeof(readbuf)) == -1) {
	if (!failed) {
	    do_log(LOGGER_ERROR, "Could not read authentication-reply from GO");
        }
	close(sockets[0]);
	sockets[0] = -1;
	failed = 1;
	return -1;
    } else {
	if (!strcmp(readbuf, "COMMAND hello,connect executed on the server!\n") ||
	    !strcmp(readbuf, "COMMAND:hello,connect -->EXECUTED on the server!\n")) {
	    do_log(LOGGER_INFO, "Logging in to GO succeeded");
	    failed = 0;
	} else {
	    if (!failed) {
		do_log(LOGGER_ERROR, "Logging in to GO failed: %s", readbuf);
	    }
	    failed = 1;
	    close(sockets[0]);
	    exit(1);
	}
    }

    // Set socket to non-blocking
    fcntl(sockets[0], F_SETFL, O_NONBLOCK);

    return 0;
}

// Check if there's incoming data from the server and send it to the right
// client. If client is gone, the data will be dropped.
void
handle_serverdata(void)
{
    char            readbuffer[BUFSIZ];
    int             bytesread;

    if ((bytesread = read(sockets[0], readbuffer, sizeof(readbuffer))) < 0) {
	if (errno == EINTR || errno == EAGAIN) {
	    return;
	} else {
	    // Read error, connection broken
	    close(sockets[0]);
	    sockets[0] = -1;
	    queue_clear();
	    do_log(LOGGER_WARNING, "Connection with GO died");
	    do_log(LOGGER_DEBUG, "Connection with GO died in handle_serverdata() at read()");
	}
    }
    // There's some data for a client
    send_serverdata(readbuffer, bytesread);

    // Scrap client from the queue
    queue_pop();
    return;
}

// Send data from whatever client to the server
void
send_clientdata(int client, char *buf, int bufsize)
{
    fd_set          writefds;
    struct timeval  tv;
    char            tempstring[480];
    char           *ptr;
    int             retval,
                    pwdlen,
                    datalen;
    int             selecttries = 0;

    // No connection with GO, can't write to server
    if (!connected())
	return;


    // Check and (if debugging is on) log request
    if (bufsize > 0) {
	// Cut off password part of the request
	pwdlen = (int)buf[0];
	if (bufsize < (pwdlen + 1)) {
	    do_log(LOGGER_WARNING, "Client %d (%s) request was too short", client, clients[client].ip);
	    return;
	}
	// Copy the request to a temporary string and log it
	datalen = buf[pwdlen + 1];
	if (bufsize < pwdlen + datalen + 2) {
	    do_log(LOGGER_WARNING, "Client %d (%s) request was too short", client, clients[client].ip);
	    return;
	}
	ptr = buf + pwdlen + 2;

	if (datalen > sizeof(tempstring) - 1) {
	    do_log(LOGGER_WARNING, "Client %d (%s) request was too long", client, clients[client].ip);
	    return;
	}

	strncpy(tempstring, ptr, datalen);
	tempstring[datalen] = '\0';
	do_log(LOGGER_DEBUG, "Client %d: %s", client, tempstring);
    }


    FD_ZERO(&writefds);
    FD_SET(sockets[0], &writefds);

    tv.tv_sec = 0;
    tv.tv_usec = 20000;

    // Try to do the select a few times
    while ((retval = select(sockets[0] + 1, NULL, &writefds, NULL, &tv)) == 0) {
	selecttries++;
	if (selecttries == 3)
	    break;
    }

    if (retval < 0 && !(errno == EINTR || errno == EAGAIN)) {
	// argh, server connect died 
	do_log(LOGGER_DEBUG, "select() failed in send_clientdata(): %s", strerror(errno));
	do_log(LOGGER_ERROR, "Connection with GO died");
	close(sockets[0]);
	sockets[0] = -1;
	queue_clear();
	return;
    } else {
	if (retval > 0 && FD_ISSET(sockets[0], &writefds)) {
	    if (write(sockets[0], buf, bufsize) == -1) {
		// Connect died
		do_log(LOGGER_DEBUG, "write() failed in send_clientdata(): %s", strerror(errno));
		do_log(LOGGER_ERROR, "Connection with GO died");
		close(sockets[0]);
		sockets[0] = -1;
		queue_clear();
		return;
	    }
	} else {
	    // The server wasn't ready for writing
	    do_log(LOGGER_DEBUG, "GO wasn't ready to receive data in send_clientdata()");
	    return;
	}
    }
    queue_put(client);		// Queue this client, if queue is too
    				// short, things screw up
}
